/*
 * Copyright 2011 Timothy Sparg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.kilnhg.timsparg.juniorkilnclient.spi;

import com.kilnhg.timsparg.juniorkilndomain.Cat;
import com.kilnhg.timsparg.juniorkilndomain.Changeset;
import com.kilnhg.timsparg.juniorkilndomain.Manifest;
import com.kilnhg.timsparg.juniorkilndomain.Project;
import com.kilnhg.timsparg.juniorkilndomain.Repository;
import com.kilnhg.timsparg.juniorkilndomain.RepositoryAlias;
import com.kilnhg.timsparg.juniorkilndomain.RepositoryGroup;
import java.net.URI;
import java.util.List;
import org.apache.http.message.BasicNameValuePair;
import com.kilnhg.timsparg.juniorkilnclient.AbstractClient;

import static com.kilnhg.timsparg.juniorkilnclient.KilnResource.*;

/**
 *
 * @author Tim Sparg
 */
public class KilnClient extends AbstractClient {

    /**
     * Constructor takes in the Kiln base URL ie https://tim-sparg.kilnhg.com
     * and the username and password.
     * 
     * 
     * @param baseURI
     * @param user
     * @param pwd
     * @throws KilnClientException 
     */
    public KilnClient(URI baseURI, String user, char[] pwd) throws KilnClientException {
        super(baseURI, "1.0");

        authenticate(user, pwd);
    }

    /**
     * Returns a list of all the project records that are readable by you. 
     * Use this method to determine the ixProjects, ixRepoGroups, 
     * and ixRepos you need.
     * @return List<Project> a list of projects.
     * @throws KilnClientException 
     */
    public List<Project> getProjects() {
        return createList(kilnGet(PROJECT_GETALL.getSnippet()), Project.class);
    }

    /**
     * You must be an administrator to create a project. Takes
     * 
     * sName: the project name, which must be unique and contain at least some 
     *        characters
     * sDescription: the project description (optional; defaults to "none")
     * sDefaultPermission: the default project permission (optional; defaults 
     *                     to "none")
     * 
     * and returns the new project's project record. The primary contact of the 
     * project will be you.
     * 
     * 
     *The default project permission determines who it is visible to and can 
     * read from it. The default project permission is also used for 
     * repositories. If a person's permission for a repository is not given, 
     * it defaults to "inherit", and the permission inherited is the default 
     * project permission of the project the repository belongs to. 
     * 
     * @param sName - the project name, which must be unique and contain at least some characters
     * @param sDescription - the project description (optional; defaults to "none")
     * @param sDefaultPermission - the default project permission (optional; defaults to "none")
     * @return the Project record
     */
    public Project createProject(String sName, String sDescription, String sDefaultPermission) {
        if (sName == null || sName.isEmpty()) {
            throw new IllegalArgumentException("sName is a requred paramter");
        }

        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("sName", sName),
                new BasicNameValuePair("sDescription", sDescription),
                new BasicNameValuePair("sDefaultPermission", sDefaultPermission));

        String returnBody = kilnPost(PROJECT_CREATE.getSnippet(), params);

        return createObject(returnBody, Project.class);
    }

    /**
     * Returns a list of all the project records that are readable by you. 
     * Use this method to determine the ixProjects, ixRepoGroups, and ixRepos 
     * you need.
     * 
     * 
     * @param ixProject
     * @return the pertinent Project
     */
    public Project getProject(int ixProject) {
        String snippet = PROJECT_GET.getSnippet().
                replace("{ixProject}", "" + ixProject);
        return createObject(kilnGet(snippet, new BasicNameValuePair("ixProject", "" + ixProject)),
                Project.class);
    }

    /**
     * You must be an administrator to write to a project. Takes
     * 
     * 
     * sName: the new project name (optional)
     * sDescription: the new project description (optional)
     * sDefaultPermission: the new default project permission (optional)
     * 
     * 
     * and updates the fields on the project corresponding to "ixProject". 
     * Returns the updated project record.
     * 
     * 
     * @param ixProject index of Project
     * @param sName new Name
     * @param sDescription new Description
     * @param sDefaultPermission new Permission
     * @return the edited Project
     */
    public Project editProject(int ixProject, String sName, String sDescription, String sDefaultPermission) {
        String snippet = PROJECT_EDIT.getSnippet().
                replace("{ixProject}", "" + ixProject);

        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("sName", sName),
                new BasicNameValuePair("sDescription", sDescription),
                new BasicNameValuePair("sDefaultPermission", sDefaultPermission));
        return createObject(kilnPost(snippet, params), Project.class);
    }

    /**
     * You must be an administrator to delete a project. Deletes the project 
     * corresponding to "ixProject". Returns true on success.
     * @param ixProject
     * @return true on deletion success
     */
    public boolean deleteProject(int ixProject) {
        String snippet = PROJECT_DELETE.getSnippet().
                replace("{ixProject}", "" + ixProject);
        return createObject(kilnPost(snippet), Boolean.class);
    }

    /**
     * You need to be a project administrator to create a repository group 
     * in the project. Takes
     *
     * sName: the repository group name
     * ixProject: the project the repository group belongs to
     *
     *and returns the new repository group's repository group record.
     * 
     * @param sName the RepositoryGroup name
     * @param ixProject the owning projects index
     * @return the newly created RepositoryGroup
     */
    public RepositoryGroup createRepositoryGroup(String sName, int ixProject) {
        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("sName", sName),
                new BasicNameValuePair("ixProject", "" + ixProject));

        return createObject(kilnPost(REPOGROUP_CREATE.getSnippet(),
                params), RepositoryGroup.class);
    }

    /**
     * You need read permissions for the project the repository group 
     * belongs to. Returns the repository group record corresponding to ixRepoGroup.
     * 
     * @param ixProjectGroup
     * @return the repositorygroup record
     */
    public RepositoryGroup getRepositoryGroup(int ixRepoGroup) {
        String snippet = REPOGROUP_GET.getSnippet().
                replace("{ixRepoGroup}", "" + ixRepoGroup);

        return createObject(kilnGet(snippet), RepositoryGroup.class);
    }

    /**
     * Takes:
     *
     * sName: the new repository group name
     *
     *and updates the fields on the repository group corresponding 
     * to "ixRepoGroup". Returns the updated repository group record.
     *
     * 
     * @param string the new name
     * @return the updated repository group record
     */
    public RepositoryGroup editRepositoryGroup(String sName, int ixRepoGroup) {
        if (sName == null || sName.isEmpty()) {
            throw new IllegalArgumentException("the new name must have a value");
        }
        String snippet = REPOGROUP_EDIT.getSnippet().
                replace("{ixRepoGroup}", "" + ixRepoGroup);
        return createObject(kilnPost(snippet,
                new BasicNameValuePair("sName", sName)), RepositoryGroup.class);
    }

    /**
     * You need to be a project administrator to delete a repository group in 
     * the project. Returns true on success.
     * 
     * @param ixRepoGroup the repositorygroups Index
     * @return true on success
     */
    public boolean deleteRepositoryGroup(int ixRepoGroup) {
        String snippet = REPOGROUP_DELETE.getSnippet().
                replace("{ixRepoGroup}", "" + ixRepoGroup);

        return createObject(kilnPost(snippet), Boolean.class);
    }

    /**
     * You need to be a project administrator to create a central repository. You need to have project write permissions to the project to create a branch repository. Takes
     *   
     *    sName (required): the new repository name
     *    sDescription (optional): the new repository description
     *    ixRepoGroup (required): the repository group the new repository will belong to
     *    ixParent (optional; if given, the ixParent repository will be branched to create the new repository): the parent repository
     *    fCentral (optional; defaults to true): true if the new repository should be central; false otherwise
     *    sDefaultPermission (optional; defaults to "none"): the default repository permission, one of "none", "read", or "modify"
     *    
     *    Returns the new repository's repository record, which does not guarantee the backend has created the repository yet. This is an asynchronous call. You should poll Api/{version}/Repo/{ixRepo} until "sStatus" becomes "good" to determine when the backend has finished. This call will be slower if the backend needs to branch a parent repository.
     *    
     *    Important note: Setting ixParent on this call is how you branch repositories through the API.
     *    
     *    @param sName    the new repository name
     *    @param sDescription the new repository description
     *    @param ixRepoGroup the repository group the new repository will belong to
     *    @param ixParent the ixParent repository will be branched to create the new repository): the parent repository
     *    @param fCentral  true if the new repository should be central; false otherwise
     *    @param sDefaultPermission the default repository permission, one of "none", "read", or "modify"
     *    @return the newly created Repository
     */
    public Repository createRepository(String sName, String sDescription,
            int ixRepoGroup, Integer ixParent, Boolean fCentral,
            String sDefaultPermission) {

        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("sName", sName),
                new BasicNameValuePair("sDescription", sDescription),
                new BasicNameValuePair("ixRepoGroup", "" + ixRepoGroup),
                new BasicNameValuePair("ixParent", ixParent == null ? "" : "" + ixParent),
                new BasicNameValuePair("fCentral", fCentral == null ? "" : "" + fCentral),
                new BasicNameValuePair("sDefaultPermission", sDefaultPermission == null ? "" : "" + sDefaultPermission));

        return createObject(kilnPost(REPO_CREATE.getSnippet(), params), Repository.class);
    }

    /**
     * You need read permissions to the repository. Returns the repository record.
     * 
     * @param ixRepo
     * @return the repository record
     */
    public Repository getRepository(int ixRepo) {
        String snippet = REPO_GET.getSnippet().
                replace("{ixRepo}", "" + ixRepo);
        return createObject(kilnGet(snippet), Repository.class);
    }

    /**
     *    You need to be the repository creator or have administrator permissions to modify a branch repository; you need to have administrator permissions to modify a central repository. Takes
     *    
     *    sName: the new repository name
     *    sDescription: the new repository description
     *    ixRepoGroup: the new repository group to move the repository 
     *    ixParent: the new repository parent (you must have read permission to the new parent)
     *    fCentral: if true, converts the branch repository into a central repository (can only be set for branch repositories)
     *    sDefaultPermission: the new default repository permission
     *    
     *    (All optional.) and updates the repository with the fields. Returns the updated repository record.
     *    
     *    
     *    @param ixRepo the new repository name
     *    @param sName the new repository description
     *    @param sDescription the new repository group to move the repository 
     *    @param ixRepoGroup the new repository parent (you must have read permission to the new parent)
     *    @param ixParent the new repository parent (you must have read permission to the new parent)
     *    @param fCentral if true, converts the branch repository into a central repository (can only be set for branch repositories)
     *    @param sDefaultPermission the new default repository permission
     *    @return the edited Repository Record
     */
    public Repository editRepository(int ixRepo, String sName, String sDescription,
            Integer ixRepoGroup, Integer ixParent, Boolean fCentral,
            String sDefaultPermission) {

        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("sName", sName),
                new BasicNameValuePair("sDescription", sDescription),
                new BasicNameValuePair("ixRepoGroup", ixRepoGroup == null ? "" : "" + ixRepoGroup),
                new BasicNameValuePair("ixParent", ixParent == null ? "" : "" + ixParent),
                new BasicNameValuePair("fCentral", fCentral == null ? "" : "" + fCentral),
                new BasicNameValuePair("sDefaultPermission", sDefaultPermission == null ? "" : "" + sDefaultPermission));

        String snippet = REPO_EDIT.getSnippet().
                replace("{ixRepo}", "" + ixRepo);

        return createObject(kilnPost(snippet, params), Repository.class);
    }

    /**
     * You need to be the repository creator or have administrator permissions 
     * to delete a branch repository; you need to have administrator permissions 
     * on the repository and the project to delete a central repository. 
     * Repositories can be undeleted by a project administrator from the 
     * website user interface.
     * 
     * @param ixRepo
     * @return true on delete success.
     */
    public boolean deleteRepository(int ixRepo) {
        String snippet = REPO_DELETE.getSnippet().
                replace("{ixRepo}", "" + ixRepo);

        return createObject(kilnPost(snippet), Boolean.class);
    }

    /**
     *    Requires read permissions to the repository. Takes
     *    
     *    ixOtherRepo: the repository to compare to
     *    revOlderThan: only include changesets older than this outgoing changeset (optional; defaults to tip)
     *    nChangesetLimit: the number of changesets to return (optional; defaults to 30)
     *    
     *    and returns a list of outgoing changeset records.    
     *    
     *    @param ixRepo base Repo
     *    @param ixOtherRepo the repository to compare to
     *    @param revOlderThan only include changesets older than this outgoing changeset (optional; defaults to tip)
     *    @param nChangesetLimit the number of changesets to return (optional; defaults to 30)
     *    @return List of changeSets
     */
    public List<Changeset> repoOutgoing(int ixRepo, int ixOtherRepo, String revOlderThan, Integer nChangesetLimit) {
        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("ixOtherRepo", "" + ixOtherRepo),
                new BasicNameValuePair("revOlderThan", revOlderThan == null ? "" : revOlderThan),
                new BasicNameValuePair("nChangesetLimit", nChangesetLimit == null ? "" : "" + nChangesetLimit));

        String snippet = REPO_OUTGOING.getSnippet().
                replace("{ixRepo}", "" + ixRepo);

        return createList(kilnGet(snippet, params), Changeset.class);
    }

    /**
     * Where ixTargetRepo is the target repository to push to. Requires read 
     * permissions to the ixRepo source repository and write permissions to to
     * the ixTargetRepo target repository. The target repository must contain a 
     * strict subset of the changesets in the source repository; the source must 
     * be related to the target; and the push must not create new heads in the 
     * target. Takes no other parameters. Returns true on success.
     * @param ixRepo
     * @param ixTargetRepo
     * @return true on success
     */
    public boolean repoPush(int ixRepo, int ixTargetRepo) {
        String snippet = REPO_PUSH.getSnippet().
                replace("{ixRepo}", "" + ixRepo).
                replace("{ixTargetRepo}", "" + ixTargetRepo);

        return createObject(kilnPost(snippet), Boolean.class);
    }

    /**
     *    Requires read permissions to the repository. Takes
     *    
     *    revOldest: revision of the oldest changeset to return (optional)
     *    nChangesetLimit: number of changesets to return, with a maximum of 100 
     *    (optional; defaults to 10)
     *    
     *    and returns a list of changeset records from the repository history, 
     *    analogous to "hg log". If revOldest is omitted, the call returns 
     *    "nChangesetLimit" records starting from the tip and working backward.
     *    
     *    @param ixRepo
     *    @return the Repository History
     */
    public List<Changeset> repoHistory(int ixRepo, String revOldest, Integer nChangesetLimit) {
        String snippet = REPO_HISTORY.getSnippet().
                replace("{ixRepo}", "" + ixRepo);

        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("revOldest", revOldest == null ? "" : revOldest),
                new BasicNameValuePair("nChangesetLimit", nChangesetLimit == null ? "" : "" + nChangesetLimit));
        return createList(kilnGet(snippet, params), Changeset.class);
    }

    /**
     *    Requires read permissions to the repository. Returns the changeset-with-diffs 
     *    record corresponding to the revision "rev" from the history, analogous to 
     *    an amalgamation "hg diff" and "hg log".
     *    
     * @param ixRepo
     * @return the Repository History
     */
    public Changeset repoHistory(int ixRepo, String rev) {
        String snippet = REPO_REVISIONHISTORY.getSnippet().
                replace("{ixRepo}", "" + ixRepo).
                replace("{rev}", rev);

        return createObject(kilnGet(snippet), Changeset.class);
    }

    /**
     *    Requires read permissions to the repository. Takes
     *    
     *    rev: the revision to read the file from (optional; default to tip)
     *    fAnnotate: if true, returns annotation records as well (optional; defaults to false)
     *    
     *    and returns the cat record corresponding to the file at "bpPath" at revision 
     *    "rev", analogous to "hg cat". Note that "bpPath" must be an encoded byte 
     *    path; see the API Idioms page for details.
     *    
     *    @param ixRepo
     *    @param bpPath
     *    @return the Cat record of the File
     */
    public Cat repoFile(int ixRepo, String bpPath) {
        String snippet = REPO_GETFILE.getSnippet().
                replace("{ixRepo}", "" + ixRepo).
                replace("{bpPath}", toHex(bpPath));

        return createObject(kilnGet(snippet), Cat.class);
    }

    /**
     *    Takes
     *    
     *    revTails: a list of revision tails
     *    
     *    and returns a list of repository records that contains a tail from 
     *    "revTails" and are readable by the user.
     *    
     *    @param revTails
     *    @return list of related repository records
     */
    public List<Repository> repoRelated(String... revTails) {
        BasicNameValuePair bnvp = new BasicNameValuePair("revTails", gson.toJson(revTails));

        return createList(kilnGet(REPO_RELATED.getSnippet(), bnvp), Repository.class);
    }

    /**
     *    Requires read permissions to the repository. Returns a list of repository 
     *    records that are related to the repository corresponding to "ixRepo" and 
     *    are readable by the user.
     *    
     *    @param ixRepo
     *    @return list of related repositories
     */
    public List<Repository> repoRelated(int ixRepo) {
        String snippet = REPO_RELATEDTOREPO.getSnippet().
                replace("{ixRepo}", "" + ixRepo);

        return createList(kilnGet(snippet), Repository.class);
    }

    /**
     *   Requires read permissions to the repository. Takes
     *    
     *    rev: the revision to read the manifest from (optional; defaults to tip)
     *    
     *    and returns a list of manifest records at revision "rev".
     *    
     *    @param ixRepo
     *    @param rev
     *    @return the Repositories manifest
     */
    public List<Manifest> repoManifest(int ixRepo, String rev) {
        String snippet = REPO_MANIFEST.getSnippet().
                replace("{ixRepo}", "" + ixRepo);

        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("rev", rev));

        return createList(kilnGet(snippet, params), Manifest.class);
    }

    /**
     * Create a repository alias with the specified name, for the given repository. You must be a project administrator to create an alias in a project.
     *
     *Takes the following parameters: 
     *
     *   ixRepo (required): the repository to attach the alias to.
     *    sName (required): the name of the alias. May contain word characters, hyphens, and parentheses (i.e. must match ^[\w- \(\)]+$)
     *
     * @param ixRepo
     * @param sName
     * @return the newly created Repository Alias
     */
    public RepositoryAlias createRepoAlias(int ixRepo, String sName) {
        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("ixRepo", "" + ixRepo),
                new BasicNameValuePair("sName", sName));

        return createObject(kilnPost(REPOALIAS_CREATE.getSnippet(), params), RepositoryAlias.class);
    }

    /**
     * Fetches the repository alias record. Requires read permissions on the repository's project.
     *
     * 
     * @param ixRepoAlias
     * @return the Repository Alias
     */
    public RepositoryAlias getRepoAlias(int ixRepoAlias) {
        String snippet = REPOALIAS_GET.getSnippet().replace("{ixRepoAlias}",
                "" + ixRepoAlias);

        return createObject(kilnGet(snippet), RepositoryAlias.class);
    }

    /**
     * Updates the specified alias. Takes either or both of the following parameters (see descriptions for Create method, above):
     *
     *  ixRepo
     *   sName
     *
     *Any parameter that is not specified remains unchanged. Requires administrator permissions on the repository's project.
     * @param ixRepoAlias
     * @param ixRepo
     * @param sName
     * @return the newly edited Repository Alias
     */
    public RepositoryAlias editRepoAlias(int ixRepoAlias, Integer ixRepo, String sName) {
        String snippet = REPOALIAS_EDIT.getSnippet().replace("{ixRepoAlias}",
                "" + ixRepoAlias);
        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("ixRepo", ixRepo == null ? "" : "" + ixRepo),
                new BasicNameValuePair("sName", sName));

        return createObject(kilnPost(snippet, params), RepositoryAlias.class);
    }

    /**
     * Deletes the specified alias. Requires administrator permissions on the associated repository's project. Returns true on success.
     * @param ixRepoAlias
     * @return true on delete success
     */
    public boolean deleteRepoAlias(int ixRepoAlias) {
        String snippet = REPOALIAS_DELETE.getSnippet().replace("{ixRepoAlias}",
                "" + ixRepoAlias);

        return createObject(kilnPost(snippet), Boolean.class);
    }

    /**
     * Returns the repository alias record by project and alias name. Takes the following parameters:
     *
     *   sProjectSlug (required): slugified project name
     *  sAliasSlug (required): slugified alias name
     *
     * 
     * @param sProjectSlug
     * @param sAliasSlug
     * @return the Repository Alias
     */
    public RepositoryAlias findRepoAlias(String sProjectSlug, String sAliasSlug) {
        BasicNameValuePair[] params = createParameters(
                new BasicNameValuePair("sProjectSlug", sProjectSlug),
                new BasicNameValuePair("sAliasSlug", sAliasSlug));

        return createObject(kilnGet(REPOALIAS_FIND.getSnippet(), params), RepositoryAlias.class);
    }
}